#include <stdio.h>
#include <jpeglib.h>
#include <setjmp.h>
#include <stdlib.h>
#include <stdint.h>
#include "jview.h"

void (*jview_error)(const char *fmt, ...);

typedef struct
{
  struct jpeg_error_mgr pub;
  jmp_buf errpos;
} jview_errman_t;


struct jview
{
  int fmt;
  int iw, ih, dw, dh, bpl, ow, oh, offset, bpp, scale;
  unsigned char *buffer;
  unsigned char **bufptr;
  jview_errman_t errman;
  struct jpeg_decompress_struct dinfo;
  unsigned char *out;
};

// These are almost all the RGB formats libjpeg knows about.
// There is another one, JCS_RGB which is equivalent to 
// JCS_EXT_RGB here.
typedef struct { const char *sv; int iv,bpp; } fmtdef_t;

static const fmtdef_t fmts[]=
{
  { "RGB",  JCS_EXT_RGB,  3 },
  { "RGBX", JCS_EXT_RGBX, 4 },
  { "BGR",  JCS_EXT_BGR,  3 },
  { "BGRX", JCS_EXT_BGRX, 4 },
  { "XBGR", JCS_EXT_XBGR, 4 },
  { "XRGB", JCS_EXT_XRGB, 4 },
  { "RGBA", JCS_EXT_RGBA, 4 },
  { "BGRA", JCS_EXT_BGRA, 4 },
  { "ABGR", JCS_EXT_ABGR, 4 },
  { "ARGB", JCS_EXT_ARGB, 4 },
  { NULL, 0 }
};


METHODDEF(void) jview_error_exit(j_common_ptr cinfo)
{
  jview_errman_t *je;
  je= (jview_errman_t*) cinfo->err;
  if (jview_error)
  {
    char msg[JMSG_LENGTH_MAX];
    (*cinfo->err->format_message)(cinfo,msg);
    (*jview_error)("libjpeg: %s\n",msg);
  }
  longjmp(je->errpos, 1);
}


jview_t* jview_init
   (const char *fmt, unsigned char *out, int dw, int dh, int bpl,
    int iw, int ih)
{
  int i;
  jview_t *jv;
  int bpp;
  for(i=0;fmts[i].sv;i++) if (!strcmp(fmt,fmts[i].sv)) break;
  if (!fmts[i].sv) return NULL;
  jv= calloc(1,sizeof(*jv));
  jv->fmt= fmts[i].iv;
  jv->bpp= fmts[i].bpp;
  jv->iw= iw; jv->ih= ih;
  jv->dw= dw; jv->dh= dh; jv->bpl= bpl; 
  if (iw>dw || ih>dh)
  {
    int xscale, yscale;
    xscale= iw/dw; yscale= ih/dh;
    if (xscale>yscale) jv->scale= xscale; else jv->scale= yscale;
    jv->ow= iw / jv->scale;
    jv->oh= ih / jv->scale;
  }
  else
  {
    jv->scale= 1;
    jv->ow= iw;
    jv->oh= ih;
  }
  jv->offset= jv->bpl*((jv->dh-jv->oh)/2) + jv->bpp*((jv->dw-jv->dw)/2);
  jv->out= out+ jv->offset;

  jv->dinfo.err = jpeg_std_error(&jv->errman.pub);
  jv->errman.pub.error_exit= jview_error_exit;
  if (setjmp(jv->errman.errpos)) goto jpegerr;
  jpeg_create_decompress(&jv->dinfo);

  if (jv->scale!=1) 
  {
    int i;
    jv->buffer= malloc(jv->iw*jv->bpp*jv->scale);
    jv->bufptr= malloc(jv->scale*sizeof(void*));
    for(i=0;i<jv->scale;i++)
      jv->bufptr[i]= jv->buffer + (jv->iw*jv->bpp)*i;
  }
  return jv;

jpegerr:
  free(jv);
  return NULL;
}

void jview_free(jview_t *jv)
{
  if (!jv) return ;
  jpeg_destroy_decompress(&jv->dinfo);
  if (jv->buffer) free(jv->buffer);
  free(jv);
}

static void jview_decode_noscale(jview_t *jv)
{
  unsigned char *out;
  unsigned char **oup;
  out= jv->out;
  oup= &out;
  while (jv->dinfo.output_scanline<jv->dinfo.output_height) 
  {
    jpeg_read_scanlines(&jv->dinfo, oup, 1);
    out += jv->bpl;
  }
}

static void jview_decode_scaled(jview_t *jv)
{
  uint32_t  *out,*p,*b;
  int i, w,r,h, scale;

  if (jv->bpp==3) return ; // FIXME: I only handle bpp==4 for now.

  out= (uint32_t*) jv->out;
  scale= jv->scale;
  r= 0;
  w= jv->ow;
  h= jv->oh;

  while (r<h)
  {
    int z;
    z= jpeg_read_scanlines(&jv->dinfo, jv->bufptr, scale);
    b= (uint32_t*) jv->buffer;
    p= out;
    for(i=0;i<w;i++) { *p= *b; p++; b+= scale; }
    out= (uint32_t*) ((unsigned char*) out + jv->bpl);
    r++;
    for(;z<scale;)
     z+= jpeg_read_scanlines(&jv->dinfo, jv->bufptr, scale-z);
  }
}

int jview_decode
  (jview_t *jv, unsigned char *inp, unsigned int inpl)
{
  if (!jv) return 1;
  if (setjmp(jv->errman.errpos)) goto jpegerr;
  jpeg_mem_src(&jv->dinfo, inp, inpl);
  jpeg_read_header(&jv->dinfo, TRUE);

  jv->dinfo.out_color_space= jv->fmt;
  jv->dinfo.scale_num= 1;
  jv->dinfo.scale_denom= 1;
  jv->dinfo.dither_mode= JDITHER_NONE;

  jpeg_start_decompress(&jv->dinfo);
  if (jv->scale==1) jview_decode_noscale(jv);
               else jview_decode_scaled(jv);
  jpeg_finish_decompress(&jv->dinfo);
  return 0;
jpegerr:
  jpeg_abort_decompress(&jv->dinfo);
  return 1;
}

