#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <linux/videodev2.h>
#include "vica.h"

void (*vica_error_output)(const char* fmt,va_list);

#define ZERO(X) memset(&(X), 0, sizeof((X)))

static  char *fourcc2str(int X)
{
  char b[5]; int i;
  for(i=0;i<4;i++) { b[i]= X; X>>= 8; }
  b[4]= 0;
  return strdup(b);
}

typedef struct
{
  int mapped;
  int index;
  int length;
  void *address;
  int deq;
} vcbuffer_t;


struct vicadev
{
  int fd;
  int last_buffer;
  int nbuffers;
  vcbuffer_t *buffers;
};

typedef struct vicadev vicadev_t;

static void* vica_error(const char *fmt, ...)
{
  va_list ap;
  if (!vica_error_output) return NULL;
  va_start(ap,fmt);
  (*vica_error_output)(fmt, ap);
  va_end(ap);
  return NULL;
}


static vica_t *close_device
  (int step, vica_t *V, int eno, const char *fmt, ...)
{
  int i;
  vicadev_t *D;
  char *es;
  va_list ap;

  D= V->dev;
  if (!vica_error_output || !fmt) goto noprint;

  if (eno) es= strerror(errno); else es= NULL;

  vica_error("[vicap %s] ", V->device_name);

  va_start(ap,fmt);
  (*vica_error_output)(fmt, ap);
  va_end(ap);

  if (es) vica_error(": %s\n", es);
  else vica_error("\n");

noprint:
  if (step<2) goto step1;
 
  for(i=0;i<D->nbuffers;i++) 
  {
    if (D->buffers[i].mapped) 
      munmap(D->buffers[i].address, D->buffers[i].length);
  }
  memset(D->buffers, 0, sizeof(D->buffers[0])*D->nbuffers);
  free(D->buffers);
step1:
  if (step<1) goto step0;
  close(D->fd);
step0:
  memset(D, 0, sizeof(D[0]));
  free(D);
  free(V->device_name);
  memset(V, 0, sizeof(V[0]));
  free(V);
  return NULL;
}

vica_t* vica_init
   (const char *devname,const char *reqfmt,int nbuffers,int reqW,int reqH)
{
  int i;
  vica_t *V;
  vicadev_t *D;
  int fmt;

  struct v4l2_capability Vcaps;
  struct v4l2_format Vfmt;
  struct v4l2_pix_format *P;
  struct v4l2_requestbuffers reqb;

  if (!devname) return vica_error("invalid device name\n");
  if (!reqfmt) return vica_error("invalid format requested\n");
  if (nbuffers<=0) return vica_error("invalid number of buffers requested\n");

  fmt= v4l2_fourcc(reqfmt[0],reqfmt[1],reqfmt[2],reqfmt[3]);

  V= calloc(1,sizeof(*V));
  D= calloc(1,sizeof(&D));
  V->device_name= strdup(devname);
  V->dev= D;

  D->fd= open(devname, O_RDWR);
  
  if (D->fd<0)
    return close_device(0, V, 1, "open() failed");

  ZERO(Vcaps);
  if (ioctl(D->fd, VIDIOC_QUERYCAP, &Vcaps))
    return close_device(1, V, 1, "querycap() failed");

#ifdef __arm__
#define device_caps capabilities
#endif
  if (!(Vcaps.device_caps & V4L2_CAP_VIDEO_CAPTURE))
    return close_device(1, V, 0, "video capture is not supported");

  if (!(Vcaps.device_caps & V4L2_CAP_STREAMING))
    return close_device(1, V, 0, "mmap() based I/O is not supported");
#ifdef __arm
#undef device_caps
#endif

  ZERO(Vfmt);
  Vfmt.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
  P= &Vfmt.fmt.pix;

  P->pixelformat= fmt;
  P->field= V4L2_FIELD_NONE;

  if (!reqW) reqW= 10000;
  if (!reqH) reqH= 10000;
  P->width= reqW;
  P->height= reqH;

  if (ioctl(D->fd,VIDIOC_S_FMT, &Vfmt)) 
    return close_device(1, V, 0, "format setting failed");
  
  V->w= P->width;
  V->h= P->height;
  V->bpl= P->bytesperline;

  if (P->field!=V4L2_FIELD_NONE)
    return close_device(1, V, 0, "non-interlaced capture is not supported");

  if (P->pixelformat!=fmt)
    return close_device(1, V, 0, "returned pixel format %s does not match "
                                 "requested format %s", 
                           fourcc2str(P->pixelformat), fourcc2str(fmt));

  ZERO(reqb);
  reqb.count= nbuffers; 
  reqb.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
  reqb.memory= V4L2_MEMORY_MMAP;

  if (ioctl(D->fd, VIDIOC_REQBUFS, &reqb))
    return close_device(1, V, 1, "buffer allocation failed");

  if (reqb.count==0) return close_device(1, V, 0, "out of device memory");

  D->nbuffers= reqb.count;
  D->buffers= calloc(D->nbuffers, sizeof(D->buffers[0]));

  for(i=0;i<D->nbuffers;i++) 
  {
    struct v4l2_buffer buf;
    ZERO(buf);
    buf.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory= V4L2_MEMORY_MMAP;
    buf.index= i;

    if (ioctl(D->fd, VIDIOC_QUERYBUF, &buf))
      return close_device(2, V, 1, "access to buffer %d denied", i);
    
    D->buffers[i].length= buf.length;
    D->buffers[i].address= mmap(NULL, buf.length, 
                                PROT_READ | PROT_WRITE , MAP_SHARED,
                                D->fd, buf.m.offset);

   if (!D->buffers[i].address) 
      return close_device(2, V, 1,
                          "mmap(offset=%d,length=%d) failed for buffer %d",
                           buf.m.offset, buf.length, i);
   D->buffers[i].deq= 1;
   D->buffers[i].length= buf.length;
   D->buffers[i].mapped= 1;
 }
 
 D->last_buffer= -1;

 return V;
}

vica_t* vica_get(vica_t *V, unsigned char **Rbuf,unsigned int *Rlen)
{
  vicadev_t *D; if (!V) return NULL; D= V->dev;
  struct v4l2_buffer buf;
  int idx;

  ZERO(buf);
  buf.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory= V4L2_MEMORY_MMAP;
  if (ioctl(D->fd, VIDIOC_DQBUF, &buf)) 
    return close_device(2, V, 1, "read_frame() failed");
  idx= buf.index;
  if (idx<0||idx>=D->nbuffers)
    return close_device(2, V, 0, "invalid buffer id %d, should be in [0,%d]",
                                  idx, D->nbuffers-1);
  *Rbuf= D->buffers[idx].address;
  if (Rlen) *Rlen= buf.bytesused;
  D->buffers[idx].deq= 1; 
  D->last_buffer= idx;
  return V;
}

int vica_fd(vica_t *V)
{
  vicadev_t *D; if (!V) return -1; D= V->dev;
  return D->fd;
}

static int putback_buffer(vicadev_t *D,int index)
{
  struct v4l2_buffer buf;
  ZERO(buf);

  buf.type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory= V4L2_MEMORY_MMAP;
  buf.index= index;

  if (ioctl(D->fd, VIDIOC_QBUF, &buf)) return 1;
  return 0;
}

vica_t* vica_put(vica_t *V)
{
  vicadev_t *D; if (!V) return NULL; D= V->dev;
  if (D->last_buffer==-1) return V;
  if (putback_buffer(D, D->last_buffer))
    return close_device(2,V, 1, "putback(%d) failed\n", D->last_buffer);
  D->last_buffer= -1;
  return V;
}

vica_t* vica_start(vica_t *V)
{
  int i;
  enum v4l2_buf_type type;
  vicadev_t *D; if (!V) return NULL; D= V->dev;

  for(i=0;i<D->nbuffers;i++) 
  {
    if (!D->buffers[i].deq) continue;
    if (putback_buffer(D,i))
      return close_device(2, V, 1, "putback(%d) failed", i);
  }

  type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if (ioctl(D->fd, VIDIOC_STREAMON, &type))
      return close_device(2, V, 1, "streamon() failed");
  return V;
}

vica_t* vica_stop(vica_t *V)
{
  vicadev_t *D; if (!V) return NULL; D= V->dev;
  enum v4l2_buf_type type;
  type= V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if (ioctl(D->fd, VIDIOC_STREAMOFF, &type)) 
      return close_device(2, V, 1, "streamoff() failed");
  return V;
}

vica_t* vica_close(vica_t *V)
{
  if (!V) return NULL;
  return close_device(2, V, 0, NULL);
}

