/**************************************************************************
  This file contains a rf24 transmitter/receiver interface. Only
  one part should be activated.

  The code is written for a simple case:
  - one transmitter
  - one receiver
  - automatic ACK without payload
  - only one pipe activated
  - IRQs enabled for transmitter and receiver MCUs
  - static payload length

 Below are the list of macros you need to define, in your config.h.

 RF24_MOSI
 RF24_MISO
 RF24_SCK
 RF24_CE
 RF24_CSN
 RF24_IRQ
   Pin definitions.

 RF24_RETRANS_DELAY
   Retransmission delay in microseconds. Valid values are 250 to 4000,
   in steps of 250 usecs.

 RF24_RETRANS_COUNT 
   Number of retransmissions before a packet is deemed lost. Can be
   0 (disabled) to 15.

 RF24_CHANNEL
   Channel number between 0 to 125

 RF24_DATA_RATE
   Choose one of RF_DR_250KBPS, RF_DR_1MBPS and RF_DR_2MBPS

 RF24_PAYLOAD_LENGTH 
   Length of data packets. The transmitter should transmit exactly
   this many bytes.

 RF24_ADDR0
 RF24_ADDR1
 RF24_ADDR2
 RF24_ADDR3
 RF24_ADDR4
   5 byte address for the receiver. ADDR0 is the lowest byte.
   If not set, the default address E7:E7:E7:E7:E7 will be used.

 RF24_TRANSMITTER
   Activates the transmitter code.
 RF24_RECEIVER
   Activates the receiver code.
 RF24_RX_CODE
   This is some user code to be executed upon the receipt of a packet
   from the radio. It's executed inside the IRQ handler. 

 Functions:
   init()    self-explanatory. 
   uint8_t send(uint8_t *B,uint8_t L)
             only the first RF24_PAYLOAD_LENGTH bytes of L are sent.
             If L is smaller, padding is added as necessary. return
             value is non-zero if no ACK was received after a number of
             retransmissions.
   uint8_t recv(uint8_t *B,uint8_t L)
             up to L bytes (a maximum of RF24_PAYLOAD_LENGTH) are
             copied from the device to the given buffer. The return
             value is the number of bytes copied.
   uint8_t have_packet()
             returns non-zero if there is a packet to be read from
             the device.
***************************************************************************/

#if defined(RF24_TRANSMITTER) || defined(RF24_RECEIVER)
#if defined(RF24_TRANSMITTER) && defined(RF24_RECEIVER)
#error "exactly one of RF24_TRANSMITTER and RF24_RECEIVER must be defined"
#endif
#else
#error "exactly one of RF24_TRANSMITTER and RF24_RECEIVER must be defined"
#endif

#define R_REGISTER          BIN(0000,0000)
#define W_REGISTER          BIN(0010,0000)
#define R_RX_PAYLOAD        BIN(0110,0001)
#define W_TX_PAYLOAD        BIN(1010,0000)
#define FLUSH_TX            BIN(1110,0001)
#define FLUSH_RX            BIN(1110,0010)
#define REUSE_TX_PL         BIN(1110,0011)
#define ACTIVATE            BIN(0101,0000)
#define R_RX_PL_WID         BIN(0110,0000)
#define W_ACK_PAYLOAD       BIN(1010,1000)
#define W_TX_PAYLOAD_NO_ACK BIN(1011,0000)
#define NOP                 BIN(1111,1111)


#define CONFIG             0x00
#define    MASK_RX_DR          (1<<6)
#define    MASK_TX_DS          (1<<5)
#define    MASK_MAX_RT         (1<<4)
#define    EN_CRC              (1<<3)
#define    CRCO_1BYTE          0
#define    CRCO_2BYTE          (1<<2)
#define    PWR_UP              (1<<1)
#define    PRIM_RX             (1<<0)
#define EN_AA              0x01
#define EN_RXADDR          0x02
#define SETUP_AW           0x03
#define    AW_3BYTES            1
#define    AW_4BYTES            2
#define    AW_5BYTES            3
#define SETUP_RETR         0x04
#define    US_TO_ARD_REAL(X)    (((X)/250-1)<<4)
#define    US_TO_ARD(X)         US_TO_ARD_REAL(MIN(4000,MAX(X,250)))
#define    ARD_TO_US(X)         ((((X)>>4)&0xf)*250+250)
#define    ARC(X)               ((X)&0xf)
#define RF_CH              0x05
#define RF_SETUP           0x06
#define    RF_DR_1MBPS          BIN(0000,0000)
#define    RF_DR_2MBPS          BIN(0000,1000)
#define    RF_DR_250KBPS        BIN(0010,0000)
#define    RF_PWR_M18dB         0
#define    RF_PWR_M12dB         (1<<1)
#define    RF_PWR_M6dB          (2<<1)
#define    RF_PWR_0dB           (3<<1)
  // this is obsolete in the plus chip
#define    LNA_HCURR            (1<<0)
#define STATUS             0x07
#define    RX_DR                (1<<6)
#define    TX_DS                (1<<5)
#define    MAX_RT               (1<<4)
#define    RX_P_NO(S)           (((S)>>1)&7)
#define    STA_TX_FULL          (1<<0)
#define OBSERVE_TX         0x08
#define    PLOS_CNT(T)          (((T)>>4)&15)
#define    ARC_CNT(T)           ((T)&15)
#define CD                 0x09
#define RX_ADDR_P0         0x0A
#define RX_ADDR_P1         0x0B
#define RX_ADDR_P2         0x0C
#define RX_ADDR_P3         0x0D
#define RX_ADDR_P4         0x0E
#define RX_ADDR_P5         0x0F
#define TX_ADDR            0x10
#define RX_PW_P0           0x11
#define RX_PW_P1           0x12
#define RX_PW_P2           0x13
#define RX_PW_P3           0x14
#define RX_PW_P4           0x15
#define RX_PW_P5           0x16
#define FIFO_STATUS        0x17
#define     TX_REUSE            (1<<6)
#define     TX_FULL             (1<<5)
#define     TX_EMPTY            (1<<4)
#define     RX_FULL             (1<<1)
#define     RX_EMPTY            (1<<0)
#define DYNPD              0x1C
#define FEATURE            0x1D
#define     EN_DPL              (1<<2)
#define     EN_ACK_PAY          (1<<1)
#define     EN_DYN_ACK          (1<<0)



/***********************************************************************
 *                                                                     *
 *                              VARIABLES                              *
 *                                                                     *
 ***********************************************************************/


static uint8_t rf24_byte(uint8_t V)
{
  uint8_t i, K;
  K= 0;
  for(i=0;i<8;i++)
  {
    if (V&0x80) PORSET( RF24_MOSI );
           else PORCLR( RF24_MOSI );
    V<<= 1;
    _delay_us(1);
    PORSET(RF24_SCK);
    _delay_us(1);
    K<<= 1;
    if (PINGET(RF24_MISO)) K|= 1;
    PORCLR(RF24_SCK);
    _delay_us(1);
  }
  return K;
}

static uint8_t rf24_buffer [ MAX(6,RF24_PAYLOAD_LENGTH) + 1];
static uint8_t rf24_len;

static void rf24_begin(uint8_t V)
{
  rf24_len= 1;
  rf24_buffer[0]= V;
}

static void rf24_put(uint8_t V)
{
  if (rf24_len>=sizeof(rf24_buffer)) return ;
  rf24_buffer[rf24_len++]= V;
}

static void rf24_end()
{
  uint8_t i;
  if (!rf24_len) return ;
  PORCLR(RF24_CSN);
  _delay_us(1);
  for(i=0;i<rf24_len;i++) rf24_buffer[i]= rf24_byte(rf24_buffer[i]);
  _delay_us(1);
  PORSET(RF24_CSN);
} 

static void rf24_wreg(uint8_t R,uint8_t V)
{ 
  rf24_begin(W_REGISTER|R); rf24_put(V); rf24_end(); 
} 


uint8_t rf24_rreg(uint8_t R)
{
  rf24_begin(R_REGISTER|R); rf24_put(0xff); rf24_end();
  return rf24_buffer[1];
}

static uint8_t rf24_single(uint8_t V)
{
  rf24_begin(V); rf24_end();
  return rf24_buffer[0]; 
}

static uint8_t rf24_status() { return rf24_single(NOP); }

void rf24_init()
{
  SETOUT(RF24_MOSI); PORCLR(RF24_MOSI); 
  SETINP(RF24_MISO); PINPU0(RF24_MISO);
  SETOUT(RF24_SCK);  PORCLR(RF24_SCK);
  SETOUT(RF24_CE);   PORCLR(RF24_CE);
  SETOUT(RF24_CSN);  PORSET(RF24_CSN);
  _delay_ms(20); 

  // the board has a 10.3 ms delay to the power-down state when
  // power is first applied.
  // from this state, we will switch on the board by setting
  // the power-up bit in the config register.
  // when we do this, the chip goes to standby-I state after
  // 1.5 ms. In this state, we can configure the chip.
  // After configuration is done, we enable the chip by setting
  // CE high. 
  // When this happens, the chip goes either to standby-II state
  // where it waits for a packet from the MCU to transmit,
  // or into the RX mode, where it waits for a packet from the
  // radio.
  // The CE pin will remain low until configuration is done.
  // after that, we will set it high and it will remain high for
  // the rest of the operation. 

  // My goal here is to make a one-way channel initially. I will
  // use only one pipe, with fixed sized packets of 32 bytes.
  // There is no way to turn off the acknowledgement mechanism, so I
  // will simply use auto-ack.

  // if the MCU did a reset but the board didn't, then we need to be 
  // on the same page. so, we flush the RX/TX buffers first.
  rf24_single(FLUSH_TX);
  rf24_single(FLUSH_RX);

#ifdef RF24_ENABLE_AUTOACK
  rf24_wreg( EN_AA, (1<<0) );
#else
  rf24_wreg( EN_AA, 0 );
#endif
#ifdef RF24_RECEIVER
  rf24_wreg( EN_RXADDR, (1<<0) );
#else
  rf24_wreg( EN_RXADDR, 0 );
#endif

#ifndef RF24_ADDRESS_WIDTH
#define RF24_ADDRESS_WIDTH 5
#endif
  rf24_wreg( SETUP_AW, RF24_ADDRESS_WIDTH-2 );

  rf24_wreg( SETUP_RETR, US_TO_ARD(RF24_RETRANS_DELAY) | RF24_RETRANS_COUNT );
//  rf24_wreg( SETUP_RETR, 0);
  rf24_wreg( RF_CH, RF24_CHANNEL );
  rf24_wreg( RF_SETUP, RF24_DATA_RATE | RF_PWR_0dB  );
  //rf24_wreg( RF_SETUP, RF24_DATA_RATE | RF_PWR_M18dB );

  // clear out the status in case there are some pending notifications
  rf24_wreg( STATUS, RX_DR | TX_DS | MAX_RT ); 
  rf24_wreg( RX_PW_P0, RF24_PAYLOAD_LENGTH );


#if defined(RF24_TRANSMITTER) && defined(RF24_ADDR0)
  rf24_begin( W_REGISTER | TX_ADDR ); 
        rf24_put(RF24_ADDR0); rf24_put(RF24_ADDR1); rf24_put(RF24_ADDR2); 
#if (RF24_ADDRESS_WIDTH>3)
        rf24_put(RF24_ADDR3);
#endif
#if (RF24_ADDRESS_WIDTH>4)
        rf24_put(RF24_ADDR4); 
#endif
  rf24_end();
#endif

  // for a transmitter, we set the P0 address to the same value
  // as the destination so that we can get ACK packets. it is
  // stupid, but I don't know what else to do here. blah.
#if defined(RF24_ADDR0) && (defined(RF24_RECEIVER) || defined(RF24_ENABLE_AUTOACK))
  rf24_begin( W_REGISTER | RX_ADDR_P0 ); 
        rf24_put(RF24_ADDR0); rf24_put(RF24_ADDR1); rf24_put(RF24_ADDR2); 
#if (RF24_ADDRESS_WIDTH>3)
        rf24_put(RF24_ADDR3);
#endif
#if (RF24_ADDRESS_WIDTH>4)
        rf24_put(RF24_ADDR4); 
#endif
  rf24_end();
#endif

  uint8_t cfg;
  // no IRQs
  cfg = MASK_RX_DR | MASK_TX_DS | MASK_MAX_RT;
  cfg|= EN_CRC | CRCO_1BYTE;
#ifdef RF24_RECEIVER
  cfg|= PRIM_RX | PWR_UP;
#else
  cfg|= PWR_UP;
#endif
  rf24_wreg(CONFIG, cfg);
  
#ifdef RF24_RECEIVER
  // for the receiver, CE is set immediately after power up.
  // for the transmitter, CE is only pulsed to start a transmission.
  PORSET(RF24_CE);
  _delay_ms(10);
#endif
}

#ifdef RF24_TRANSMITTER
void rf24_send(uint8_t *B,uint8_t L)
{
  uint8_t i,Z;

  // rf24_wreg( CONFIG, rf24_rreg(CONFIG) | PWR_UP ); _delay_us(150);

  rf24_single(FLUSH_TX);
  rf24_wreg(STATUS, rf24_status()); // clear the stuff

  rf24_begin(W_TX_PAYLOAD);
  L= MIN(L,RF24_PAYLOAD_LENGTH);
  for(i=0;i<L;i++) rf24_put(B[i]);
  for(;i<RF24_PAYLOAD_LENGTH;i++) rf24_put(0xc4);
  rf24_end();

  PORSET(RF24_CE);
  _delay_us(15);
  PORCLR(RF24_CE);

  while(1)
  {
    _delay_us(100);
    Z= rf24_status();
    // if (Z&TX_DS) break;
    if (Z&(TX_DS|MAX_RT)) break;
  } 
  // rf24_wreg(STATUS,Z | TX_DS | MAX_RT);
  rf24_wreg(STATUS, Z);
  // rf24_wreg( CONFIG, rf24_rreg(CONFIG) & ~PWR_UP );
}
#endif

#ifdef RF24_RECEIVER

uint8_t rf24_recv(uint8_t *B, uint8_t L)
{
  uint8_t i;
  L= MIN(L,RF24_PAYLOAD_LENGTH);
  rf24_begin(R_RX_PAYLOAD);
  for(i=0;i<RF24_PAYLOAD_LENGTH;i++) rf24_put(0xFF);
  rf24_end();
  for(i=0;i<L;i++) B[i]= rf24_buffer[i+1];
  rf24_wreg(STATUS, rf24_status());
//  rf24_single(FLUSH_RX);
  return L;
}

uint8_t rf24_have_packet() 
{
 return (rf24_status()&BIN(0000,1110))!=BIN(0000,1110);
//  return (rf24_status() & RX_DR)!=0;
}

#endif
